"""
Authors: Peter Mawhorter
Consulted:
Date: 2022-4-4
Purpose: Tests for the checklist task. Also runs testing functions
    defined in that file where students are required to write their own
    tests.
"""

import optimism

import checklist


#-----------------------------#
# Camping Checklist Functions #
#-----------------------------#

def makeCampingList():
    """
    Uses addTask and addList to assemble a blank camping checklist, where
    every task is unfinished and each list is empty. Returns the
    resulting checklist.
    """
    # First, we use addTask, addList, and addItemToList to build up a
    # list of camping equipment and supplies
    result = {}

    result = checklist.addTask(result, "Tent")
    result = checklist.addTask(result, "Sleeping bags")
    result = checklist.addTask(result, "Cooking supplies")
    result = checklist.addTask(result, "Camp chairs")
    result = checklist.addTask(result, "Snacks")
    result = checklist.addTask(result, "Meals")

    return result


def completeSomeOfCampingList(campingList):
    """
    Uses completeTask to fill out some of the items in a camping list
    created by `makeCampingList`.
    """
    campingList = checklist.completeTask(campingList, "Tent")
    campingList = checklist.completeTask(campingList, "Sleeping bags")
    campingList = checklist.completeTask(campingList, "Cooking supplies")
    campingList = checklist.completeTask(campingList, "Meals")

    return campingList


def testCamping():
    """
    Tests all of the different checklist functions at a high level by
    creating a checklist, completing some of it, completing all of it,
    and then resetting it. At each step we test the displayed checklist,
    as well as the status of a few tasks.
    """
    # Skip testing if required function(s) aren't defined yet
    if (
        not hasattr(checklist, "addTask")
     or not hasattr(checklist, "taskStatus")
     or not hasattr(checklist, "completeTask")
     or not hasattr(checklist, "resetChecklist")
    ):
        print(
            "Skipping camping test because one or more required"
            " functions haven't been defined yet."
        )
        return

    # Create an empty camping list
    tasks = makeCampingList()

    # Create two test managers: one for testing the displayed checklist,
    # and another for testing taskStatus results.
    testDisplay = optimism.testFunctionMaybe(checklist, "displayChecklist")
    testStatus = optimism.testFunctionMaybe(checklist, "taskStatus")

    # Display the checklist and check what gets printed
    testDisplay.case(tasks).checkPrintedLines(
        "[ ] Tent",
        "[ ] Meals",
        "[ ] Snacks",
        "[ ] Camp chairs",
        "[ ] Sleeping bags",
        "[ ] Cooking supplies",
    )
    testStatus.case(tasks, 'Tent').checkReturnValue(False)
    testStatus.case(tasks, 'Sleeping bags').checkReturnValue(False)
    testStatus.case(tasks, 'Microwave').checkReturnValue(None)

    # Fill it up with items and complete the tasks
    tasks = completeSomeOfCampingList(tasks)

    # Display the checklist and check what gets printed again
    testDisplay.case(tasks).checkPrintedLines(
        "[X] Tent",
        "[X] Meals",
        "[ ] Snacks",
        "[ ] Camp chairs",
        "[X] Sleeping bags",
        "[X] Cooking supplies",
    )
    testStatus.case(tasks, 'Tent').checkReturnValue(True)
    testStatus.case(tasks, 'Snacks').checkReturnValue(False)
    testStatus.case(tasks, 'Microwave').checkReturnValue(None)

    tasks = checklist.completeTask(tasks, 'Snacks')
    tasks = checklist.completeTask(tasks, 'Camp chairs')

    # Check full completion
    testDisplay.case(tasks).checkPrintedLines(
        "[X] Tent",
        "[X] Meals",
        "[X] Snacks",
        "[X] Camp chairs",
        "[X] Sleeping bags",
        "[X] Cooking supplies",
    )
    testStatus.case(tasks, 'Tent').checkReturnValue(True)
    testStatus.case(tasks, 'Camp chairs').checkReturnValue(True)
    testStatus.case(tasks, 'Microwave').checkReturnValue(None)

    tasks = checklist.resetChecklist(tasks)

    # Check reset
    testDisplay.case(tasks).checkPrintedLines(
        "[ ] Tent",
        "[ ] Meals",
        "[ ] Snacks",
        "[ ] Camp chairs",
        "[ ] Sleeping bags",
        "[ ] Cooking supplies",
    )
    testStatus.case(tasks, 'Tent').checkReturnValue(False)
    testStatus.case(tasks, 'Camp chairs').checkReturnValue(False)
    testStatus.case(tasks, 'Microwave').checkReturnValue(None)

    tasks = completeSomeOfCampingList(tasks)

    # Display the checklist and check what gets printed again
    testDisplay.case(tasks).checkPrintedLines(
        "[X] Tent",
        "[X] Meals",
        "[ ] Snacks",
        "[ ] Camp chairs",
        "[X] Sleeping bags",
        "[X] Cooking supplies",
    )
    testStatus.case(tasks, 'Tent').checkReturnValue(True)
    testStatus.case(tasks, 'Camp chairs').checkReturnValue(False)
    testStatus.case(tasks, 'Microwave').checkReturnValue(None)


#-------------------#
# Testing functions #
#-------------------#

def testAddTask():
    """
    Tests the addTask function using optimism. The addTask function is
    supposed to add a task to the checklist with an initial value of
    False. If you try to add a task that's already in the checklist, it
    should print a message and return None instead of returning the
    modified checklist. Note that this code doesn't check the printed
    message in that case (and your tests don't have to do that either).
    """
    tester = optimism.testFunctionMaybe(checklist, 'addTask')
    tester.case({}, 'hi').checkReturnValue({'hi': False})
    tester.case({}, 'bye').checkReturnValue({'bye': False})
    tester.case({'hi': True}, 'bye').checkReturnValue(
        {'hi': True, 'bye': False}
    )
    tester.case({'hi': True}, 'hi').checkReturnValue(None)


def testTaskStatus():
    """
    Tests the taskStatus function using optimism.
    """
    tester = optimism.testFunctionMaybe(checklist, 'taskStatus')
    tester.case({'hi': True}, 'hi').checkReturnValue(True)
    tester.case({'hi': False}, 'hi').checkReturnValue(False)
    tester.case({'hi': True, 'bye': False}, 'hi').checkReturnValue(True)
    tester.case({'hi': True, 'bye': False}, 'bye').checkReturnValue(False)
    tester.case({}, 'hi').checkReturnValue(None)


def testCompleteTask():
    """
    Tests the completeTask function using optimism.
    """
    tester = optimism.testFunctionMaybe(checklist, 'completeTask')
    tester.case({'hi': False}, 'hi').checkReturnValue({'hi': True})
    tester.case({'hi': True}, 'hi').checkReturnValue({'hi': True})
    tester.case({'hi': True}, 'bye').checkReturnValue(None)
    tester.case({'hi': True, 'bye': False}, 'hi').checkReturnValue(
        {'hi': True, 'bye': False}
    )
    tester.case({'hi': True, 'bye': False}, 'bye').checkReturnValue(
        {'hi': True, 'bye': True}
    )


def testResetChecklist():
    """
    Tests the resetChecklist function using optimism.
    """
    tester = optimism.testFunctionMaybe(checklist, 'resetChecklist')
    tester.case({'hi': False}).checkReturnValue({'hi': False})
    tester.case({'hi': True}).checkReturnValue({'hi': False})
    tester.case({'hi': True, 'bye': True}).checkReturnValue(
        {'hi': False, 'bye': False}
    )
    tester.case({'hi': False, 'bye': True}).checkReturnValue(
        {'hi': False, 'bye': False}
    )
    tester.case({'hi': False, 'bye': False}).checkReturnValue(
        {'hi': False, 'bye': False}
    )


def testLengthAndValue():
    """
    Tests the lengthAndValue function using optimism.
    """
    tester = optimism.testFunctionMaybe(checklist, 'lengthAndValue')
    tester.case('hi').checkReturnValue((2, 'hi'))
    tester.case('bye').checkReturnValue((3, 'bye'))
    tester.case('soliloquy').checkReturnValue((9, 'soliloquy'))


def testDisplayChecklist():
    """
    Tests the displayChecklist function, which needs to test printed
    output.
    """
    testDC = optimism.testFunctionMaybe(checklist, "displayChecklist")
    # Basic tests w/ two entries
    testDC.case({'a': True, 'b': False}).checkPrintedLines(
        '[X] a',
        '[ ] b',
    )
    # This case tests ordering in a few ways
    testDC.case(
        {'one': True, 'two': False, 'three': False, 'z': True}
    ).checkPrintedLines(
        '[X] z',
        '[X] one',
        '[ ] two',
        '[ ] three',
    )
    # Mixed ordering of checkboxes again
    testDC.case(
        { 'abc': True, 'z': True, 'ms': False, 'mm': False, 'aaaa': True}
    ).checkPrintedLines(
        '[X] z',
        '[ ] mm',
        '[ ] ms',
        '[X] abc',
        '[X] aaaa',
    )


def testSaveChecklist():
    """
    Tests the saveChecklist function.
    """
    testSC = optimism.testFunctionMaybe(checklist, "saveChecklist")

    # Basic test w/ two entries
    testSC.case({'a': True, 'b': False}, 'ab.txt').checkFileLines(
        'ab.txt',
        '[X] a',
        '[ ] b',
    )

    # This case tests ordering in a few ways
    testSC.case(
        {'one': True, 'two': False, 'three': False, 'z': True},
        "oneTwoThree.txt"
    ).checkFileLines(
        "oneTwoThree.txt",
        '[X] z',
        '[X] one',
        '[ ] two',
        '[ ] three',
    )

    # Mixed ordering of checkboxes again
    testSC.case(
        { 'abc': True, 'z': True, 'ms': False, 'mm': False, 'aaaa': True},
        "abc.txt"
    ).checkFileLines(
        "abc.txt",
        '[X] z',
        '[ ] mm',
        '[ ] ms',
        '[X] abc',
        '[X] aaaa',
    )


def testLoadChecklist():
    """
    Tests the loadChecklist function.
    """
    testLC = optimism.testFunctionMaybe(checklist, "loadChecklist")

    with open("abLoad.txt", 'w') as fileOutput:
        fileOutput.write("[X] a\n[ ] b\n")

    # Basic test w/ two entries
    testLC.case('abLoad.txt').checkReturnValue({'a': True, 'b': False})

    with open("oneTwoThreeLoad.txt", 'w') as fileOutput:
        fileOutput.write("[X] z\n[X] one\n[ ] two\n[ ] three\n")

    # This case tests ordering in a few ways
    testLC.case("oneTwoThreeLoad.txt").checkReturnValue(
        {'one': True, 'two': False, 'three': False, 'z': True}
    )

    # Mixed ordering of checkboxes again
    with open("abcLoad.txt", 'w') as fileOutput:
        fileOutput.write("[X] z\n[ ] mm\n[ ] ms\n[X] abc\n[X] aaaa\n")

    testLC.case("abcLoad.txt").checkReturnValue(
        { 'abc': True, 'z': True, 'ms': False, 'mm': False, 'aaaa': True}
    )


#----------------------------------#
# Run all of the testing functions #
#----------------------------------#

def runTests():
    """
    Calls each required testing function from the checklist.py file.
    """
    testAddTask()
    testTaskStatus()
    testCompleteTask()
    testResetChecklist()
    testLengthAndValue()
    testDisplayChecklist()
    testSaveChecklist()
    testLoadChecklist()
    testCamping()


# This condition is only true when we're running this file (as opposed to
# importing it)
if __name__ == "__main__":
    runTests()
